#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "http_fetcher.h"

const char* argv0 = "<unknown>";
const char* file  = 0;
const char* url   = 0;
int fd    = -1;
int debug =  0;
int sim   =  0;

void updateStatistics(int bytesRead, int contentLength);
void lto_progress();
int sim_fetch(const char* url, char** fileBuf);

// const char* default_url  = "http://gllto.glpals.com/7day/v2/latest/lto2.dat";
const char* default_url  = "http://66.35.236.20/7day/v2/latest/lto2.dat";
const char* default_file = "./lto2.dat";

//-----------------------------------------------------------------------------
//
//      lto_get()
//
//  USAGE:
//      lto_get -default [+DEBUG]
//      lto_get <url> <file> [+DEBUG]
//
//  DESCRPIPTION:
//      copy file from <url> to <file>
//      Report to stdout:
//          <done>=N        When done.  N is the file size in bytes.
//                          Exit 0.
//
//          <percent>=0..100
//                          During the download itself.
//
//          <progress>=A..Z,a..z
//                          During the download itself.
//
//          <err>="...."    Upon error.
//                          Exit 1.
//
//-----------------------------------------------------------------------------

int
main(int argc, char** argv)
{
    int i = 0;

    argv0 = argv[i];

    for (++i; i < argc; ++i)
    {
        const char* a = argv[i];

        if (!strcmp(a, "-help"))
        {
            printf( "<help>=\"\n"
                    "lto_get version 0.04 %s\n"
                    "lto_get <url> <file> [opts]\n"
                    "lto_get -default [opts]\n"
                    "   defaults are: url  %s\n"
                    "                 file %s\n"
                    "Options:\n"
                    "    +DEBUG       enable more output\n"
                    "    +SIM         simulate download and file creation\n"
                    "\n",
                            __DATE__,
                            default_url, default_file);
            return 1;
        }

        if (!strcmp(a, "-default"))
        {
            url  = default_url;
            file = default_file;
            continue;
        }

        if (!strcmp(a,"+DEBUG"))
        {
            debug = 1;
            continue;
        }

        if (!strcmp(a,"+SIM"))
        {
            sim = 1;
            continue;
        }

        if (!url)
        {
            url = a;
            continue;
        }
        if (!file)
        {
            file = a;
            continue;
        }
    }

    if (debug)
    {
        printf("<debug>=\"url  %s\"\n", (url)  ? url  : "<null>" );
        printf("<debug>=\"file %s\"\n", (file) ? file : "<null>" );
    }

    if (!url || !file)
    {
        printf("<err>=\"%s missing parameters\"\n", argv0);
        return 1;
    }

    lto_progress();

    fd = open(file, O_WRONLY | O_CREAT,
                        S_IRUSR | S_IWUSR |
                        S_IRGRP | S_IWGRP |
                        S_IROTH | S_IWOTH   );
    if (fd < 0)
    {
        printf("<err>=\"%s: open(%s) error %d %s\"\n", argv0, file,
                                        errno, strerror(errno));
        return 2;
    }

    lto_progress();

    http_setTimeout(10);                // 10 seconds between packets max.
{
    char* fileBuf;
    int   write_size = 0;
    int   http_size;

    if (sim)
    {
        http_size = sim_fetch(url, &fileBuf);
    }
    else
    {
        http_size = http_fetch(url, &fileBuf);
    }

    if (http_size <= 0)
    {
        printf("<err>=\"%s: http get(%s) error %s\"\n", argv0, url,
                                        http_strerror());
        close(fd);
        return 2;
    }
    lto_progress();

    write_size = write(fd, fileBuf, http_size);
    lto_progress();
    close(fd);
    lto_progress();
    if (write_size != http_size)
    {
        printf("<err>=\"%s: write(%s,%d) error %d %s\"\n",
                                        argv0, file, http_size,
                                        errno, strerror(errno));
        return 2;
    }
    printf("<done>=%d\n", write_size);

    return 0;
}
}

//-----------------------------------------------------------------------------
//
//            updateStatistics()
//
//  DESCRPIPTION:
//      Report data transfer percentage if we've made more than 1.9%
//      progress.
//
//      Maintain the download statistics in terms of percent*10
//
//-----------------------------------------------------------------------------

void
updateStatistics(int bytesRead, int contentLength)
{
    static int last_percent_10 = -100;
    int percent_10 = ((bytesRead * 1000) + 5) / contentLength;

    // if (debug)
    // {
    //     printf("<debug>=\"percent_10(%d/%d = %d\"\n",
    //                     bytesRead, contentLength, percent_10);
    // }

    if (((percent_10 - last_percent_10) > 19) || debug)
    {
        last_percent_10 = percent_10;
        printf("<percent>=%d\n", percent_10/10);
        fflush(stdout);
    }
}

//-----------------------------------------------------------------------------
//
//            lto_progress()
//
//  DESCRPIPTION:
//      Report the progress "A" to "Z". then "a" to "z" (if needed)
//
//-----------------------------------------------------------------------------

void
lto_progress()
{
    static char progress = 'a';
    if (debug)
    {
        printf("<progress>=%c\n", progress);
        fflush(stdout);
    }
    if (++progress == ('Z' + 1))
    {
        progress = 'a';
    }
    else if (progress == ('z' + 1))
    {
        progress = 'A';
    }
}

//-----------------------------------------------------------------------------
//
//            sim_fetch()
//
//  DESCRPIPTION:
//      Simulate the downloading of the code by sleep and dummy data
//
//-----------------------------------------------------------------------------

int
sim_fetch(const char* url, char** fileBuf)
{
    const int fileSize = 10000;
    const int nLoops   = 18;
    const int loopSize = fileSize/nLoops;
    int written = 0;
    int loop;

    char* data = (char*) malloc(fileSize);

    if (debug)
    {
        printf("<debug>=\"simulating file size %d in %d loops\"\n",
                    fileSize, nLoops);
    }

    for (loop = 0; loop <= nLoops; ++loop)
    {
        updateStatistics(written, fileSize);
        sleep(1);
        written += loopSize;
        if (written > fileSize)
        {
            written = fileSize;
            updateStatistics(written, fileSize);
            break;
        }
    }

    *fileBuf = data;
    return fileSize;
}
