#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <memory.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

static int read_output(int fd);

//-----------------------------------------------------------------------------
//
//      lto_util()
//
//  USAGE:
//      lto_util
//
//  DESCRPIPTION:
//      Test the lto_get program.
//      Fork and exec lto_get() then
//      track its progress and errors,
//      then wait for it to exit.
//
//-----------------------------------------------------------------------------

int
main(int argc, char** argv)
{
    int fd[2];
    int err = pipe(fd);
    int pid = 0;

    const char* lto_get = "./lto_get";
    const char* url     = "http://gllto.glpals.com/7day/v2/latest/lto2.dat";
    const char* file    = "./lto2.dat";
    const char* debug   = "+DEBUG";

    struct stat st;

    //  Make sure that lto_get is an executable.
 
    if (-1 == lstat(lto_get, &st))
    {

        printf("lto_util: lstat(%s) Error %d %s\n",
                            lto_get, errno, strerror(errno));
        return 1;
    }

    if (!(st.st_mode & (S_IXOTH | S_IXGRP | S_IXUSR)))
    {
        printf("lto_util: \"%s\" %x mode & %x execute = %x not executable\n",
                    lto_get, st.st_mode,
                    S_IXOTH | S_IXGRP | S_IXUSR,
                    st.st_mode & (S_IXOTH | S_IXGRP | S_IXUSR));
        return 1;
    }

    if (err)
    {
        printf("pipe() err %d %s\n", errno, strerror(errno));
        return 1;
    }

    pid = fork();
    if (pid)            // we're the parent - wait for the child.
    {
        pid_t x;

        close(fd[1]);       // close the write side

        err = read_output(fd[0]);
        x = waitpid(pid, 0, 0);
        if (x != pid)
        {
            printf("%d = waitpid(%d) err %d %s\n",
                    x, pid, errno, strerror(errno));
        }
    }
    else
    {
        int fd1;
        int fd2;
        // we're the child - exec the program.

        // close(0);        // close stdin
        close(1);           // close stdout
        close(fd[0]);       // close the write side

        fd1 = dup(fd[1]);   // reopen stdout and stderr
        if (fd1 != 1)
        {
            fprintf(stderr,"<err>=failed to reopen stdout %d %s\n",
                        errno, strerror(errno));
            return 3;
        }

        close(2);           // close stderr
        fd2 = dup(fd[1]);   // to point to our pipe.
        if (fd2 != 2)
        {
            printf("<err>=failed to reopen stderr %d %s\n",
                        errno, strerror(errno));
            return 3;
        }

        printf("<progress>=Z\n");       // indicate we're alive

        err = execl(lto_get, lto_get, url, file, "+DEBUG", (char *) NULL);
        printf("<err>=\"execl(%s,%s,%s,%s) error %d %s\"\n",
                    lto_get, url, file, debug,
                    errno, strerror(errno));
        return 3;
    }

    return 0;
}

static int
parse_output(const char* buf)
{
    // printf("parse_output(%s) %d\n", buf, strlen(buf));
    if (!strncmp(buf, "<done>=", 7))
    {
        int n;

        sscanf(buf + 7, "%d", &n);
        printf("done - %d bytes\n", n);
        return 0;
    }

    if (!strncmp(buf, "<percent>=", 10))
    {
        int n;

        sscanf(buf + 10, "%d", &n);
        printf("percent =  %d%%\n", n);
        return -1;
    }

    if (!strncmp(buf, "<progress>=", 11))
    {
        char progress = buf[11];
        printf("progress = '%c'\n", progress);
        return -2;
    }

    if (!strncmp(buf, "<err>=", 6))
    {
        const char* err = buf + 6;
        printf("error = %s\n", err);
        return 1;
    }

    if (!strncmp(buf, "<debug>=", 8))
    {
        const char* dbg = buf + 8;
        printf("debug = %s\n", dbg);
        return -3;
    }

    printf("Unparsed output \"%s\"\n", buf);
    return 2;
}

static int
read_output(int fd)
{
    int offset = 0;
    while (1)
    {
        char* p;
        char* pNext;
        char  buf[1024];
        int n = read(fd, buf + offset, sizeof(buf) - offset);
        int err;
        if (n < 0)
        {
            printf("read(%d) error %d %s\n", n, errno, strerror(errno));
            return 2;
        }
        else if (n > 0)
        {
            buf[n] = 0;
            // printf("read(%d,%d,%s)\n", fd, n, buf);
        }
        buf[n] = 0;

        for (p = buf; p < buf + n; p = pNext + 1)
        {
            pNext = strchr(p, '\n');
            if (!pNext)
            {
                pNext = buf+n;
                offset = pNext - p;
                memcpy(buf, p, offset);
                buf[offset] = 0;
                // printf("Offset %d\n", offset);
                break;
            }
            *pNext = 0;
            err = parse_output(p);

            if (err >= 0)
            {
                return err;
            }
        }
    }
}
