/*
 * Copyright © 2007 Global Locate.
 *
 * Written by Ken Yale.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "lto_download.h"
#include "liblto.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <memory.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <gtk/gtk.h>

//      See page_LTO.cpp for these:
extern void lto_download_error(const char* msg);
extern void lto_download_progress(char progress);
extern void lto_download_percent(int n);
extern void lto_download_done(void);

static gint lto_cb_tag =  0;
static int  lto_fd     = -1;
static int  lto_pid    =  0;
static bool lto_downloading_flag = false;

static void lto_wait_for_stop(void);
static void lto_fd_read(gpointer data, gint source, GdkInputCondition cond);
static const char* ltoDownloadPidWait();
static void lto_read_init();

const char* cc_lto_get = "/home/root/gpsd/lto_get";
// st char* cc_url     = "http://66.35.236.20"  "/7day/v2/latest/lto2.dat";
const char* cc_url     = "http://gllto.glpals.com/7day/v2/latest/lto2.dat";
const char* cc_file    = "/home/root/gpsd/lto2.dat";

//-----------------------------------------------------------------------------------
//
//                      ltoRemove()
//
//-----------------------------------------------------------------------------------

const char*
ltoRemove()
{
    int err = unlink(cc_file);
    if (-1 == err)
    {
        return strerror(errno);
    }
    return 0;
}

//-----------------------------------------------------------------------------------
//
//                      ltoLastDownloaded()
//
//-----------------------------------------------------------------------------------

int
ltoLastLoaded(int n, char* msg)
{
    struct stat lto_stat;
    if (-1 == stat(cc_file, &lto_stat))
    {
        return sprintf(msg, "Can't find %s", cc_file);
    }

    struct tm* p = localtime(&lto_stat.st_mtime);
    return strftime(msg, n, "LTO last loaded %x %X", p);
}

//------------------------------------------------------------------------------------
//
//                      ltoDownloadStart()
//
//------------------------------------------------------------------------------------

const char*
ltoDownloadStart()
{
    if (lto_downloading_flag)
    {
        return "Download in progress";
    }
    lto_read_init();

    const char* lto_get = cc_lto_get;
    const char* url     = cc_url;
    const char* file    = cc_file;
    const char* debug   = "+DEBUG";
    // const char* sim     = "+SIM";
    const char* sim     = 0;

#if 1
    printf("ltoDownloadStart()/%d URL %s  FILE %s  PROGRAM %s  %s\n",
                            __LINE__, url, file, lto_get, debug);
#endif

    //  Make sure that lto_get is an executable.

    struct stat st;
    if (-1 == lstat(lto_get, &st))
    {
        fprintf(stderr, "omgui: lstat(%s) Error %d %s\n",
                            lto_get, errno, strerror(errno));
        return "Can't find lto_get program";
    }

    if (!(st.st_mode & (S_IXOTH | S_IXGRP | S_IXUSR)))
    {
        fprintf(stderr, "omgui: \"%s\" mode %o not executable\n",
                    lto_get, st.st_mode);
        return "lto_program is not executable";
    }

    int fd[2];
    int err = pipe(fd);
    if (err)
    {
        fprintf(stderr, "omgui: pipe() err %d %s\n", errno, strerror(errno));
        return "Can't open pipe";
    }

    int pid = fork();
    if (pid)
    {
        //--------------------------------------------------
        //      We're the parent - connect child output
        //      to GTK+ and then wait for input.
        //--------------------------------------------------

        close(fd[1]);                   // close the write side
        lto_fd  = fd[0];
        lto_pid = pid;
        lto_cb_tag = gdk_input_add(lto_fd, GDK_INPUT_READ, lto_fd_read, 0);
        return 0;
    }

    //--------------------------------------------------
    // we're the child - exec the program.
    //--------------------------------------------------

#if 1
    printf("<debug>=\"starting to execl(%s,%s,%s,%s,%s)\"\n",
                            lto_get, url, file, debug, sim);

    int opt = fcntl(fd[1], F_GETFL, 0);
    if (-1 == opt)
    {
        fprintf(stderr,"<err>=fcntl(%d,F_GETFL,0) error %d %s\"\n",
                    fd[1], errno, strerror(errno));
    }
    else
    {
        opt |= O_NONBLOCK;
        opt = fcntl(fd[1], F_SETFL, opt);
        if (-1 == opt)
        {
            fprintf(stderr,"<err>=fcntl(%d,F_SETFL,0x%x) error %d %s\"\n",
                        fd[1], opt, errno, strerror(errno));
        }
    }
#endif

    // close(0);        // close stdin
    close(1);           // close stdout
    close(fd[0]);       // close the write side

    int fd1 = dup(fd[1]);   // reopen stdout and stderr
    if (fd1 != 1)
    {
        fprintf(stderr,"<err>=failed to reopen stdout %d %s\n",
                    errno, strerror(errno));
        exit(3);
    }

    close(2);               // close stderr
    int fd2 = dup(fd[1]);   // to point to our pipe.
    if (fd2 != 2)
    {
        fprintf(stderr, "<err>=failed to reopen stderr %d %s\n",
                    errno, strerror(errno));
        exit(3);
    }

    printf("<progress>=a\n");       // indicate we're alive

    err = execl(lto_get, lto_get, url, file, debug, sim, (char *) NULL);
    printf("<err>=\"execl(%s,%s,%s,%s,%s) error %d %s\"\n",
                lto_get, url, file, debug, sim,
                errno, strerror(errno));
    exit(3);
    return "fork() and exec() failed";
}

//-----------------------------------------------------------------------------
//
//                      lto_parse_output()
//
//      Decode the lto_get process output.
//
//      Returns:
//              0       Done.  Successfull.
//              n < 0   Normal progress report.  Continue.
//              n > 0   Error code.  Should stop.
//
//-----------------------------------------------------------------------------

static int
lto_parse_output(const char* buf)
{
    // printf("parse_output(%s) %d\n", buf, strlen(buf));
    if (!strncmp(buf, "<done>=", 7))
    {
        int n;

        sscanf(buf + 7, "%d", &n);
        // printf("done - %d bytes\n", n);
        return 0;
    }

    if (!strncmp(buf, "<percent>=", 10))
    {
        int n;

        sscanf(buf + 10, "%d", &n);
        // printf("percent =  %d%%\n", n);
        lto_download_percent(n);
        return -1;
    }

    if (!strncmp(buf, "<progress>=", 11))
    {
        char progress = buf[11];
        // printf("progress = '%c'\n", progress);
        lto_download_progress(progress);
        return -2;
    }

    if (!strncmp(buf, "<err>=", 6))
    {
        const char* err = buf + 6;
        // printf("error = %s\n", err);
        lto_download_error(err);
        return 1;
    }

    if (!strncmp(buf, "<debug>=", 8))
    {
        const char* dbg = buf + 8;
        printf("debug = %s\n", dbg);
        return -3;
    }

    printf("Unparsed lto output \"%s\"\n", buf);
    return 2;
}

//-----------------------------------------------------------------------------
//
//              lto_read_output()
//
//      Read output of the lto_get process.
//
//      Returns:
//              0       Done.  Successfull.
//              n < 0   Normal progress report.  Continue.
//              n > 0   Error code.  Should stop.
//
//-----------------------------------------------------------------------------

static int  offset;
static char buf[1024];

static void
lto_read_init()
{
    offset = 0;
    buf[0] = 0;
}

static int
lto_read_output(int fd)
{
    // printf("lto_read(%d,buf,%d)\n", fd, sizeof(buf) - offset); 
    int n = read(fd, buf + offset, sizeof(buf) - offset);
    if (n < 0)
    {
        // printf("lto_read(%d) error %d %s\n", n, errno, strerror(errno));
        return 2;
    }
    else if (n > 0)
    {
        buf[n] = 0;
        // printf("lto_read(%d,%d,%s)\n", fd, n, buf);
    }
    else
    {
        // printf("read(%d,%d)\n", fd, n);
        return -4;
    }
    char* pEnd = buf + n + offset;
    *pEnd = 0;

    char* pNext;
    for (char* p = buf; p < pEnd; p = pNext + 1)
    {
        pNext = strchr(p, '\n');
        if (!pNext)
        {
            pNext  = pEnd;
            offset = pNext - p;                 // printf("Offset %d\n", offset);
            memcpy(buf, p, offset);
            buf[offset] = 0;
            break;
        }
        *pNext = 0;

        int err = lto_parse_output(p);
        if (err >= 0)
        {
            return err;
        }
    }
    return -5;                  // Incomplete read.
}

static void
lto_fd_read(gpointer data, gint source, GdkInputCondition cond)
{
    if (source != lto_fd)
    {
        lto_download_error("fd mismatch");
        return;
    }
    if (GDK_INPUT_READ != cond)
    {
        lto_download_error("gtk read callback condition mismatch");
        return;
    }
    int err = lto_read_output(lto_fd);

    if (err >= 0)
    {
        // printf("omgui: lto_status %d\n", err);
        // fflush(stdout);
        lto_wait_for_stop();
    }
}

//-----------------------------------------------------------------------------
//
//              lto_download_stop()
//
//      Disconnect the LTO_GET process from GTK+
//
//-----------------------------------------------------------------------------

static void
lto_wait_for_stop()
{
    const char* msg = ltoDownloadPidWait();
    if (msg)
    {
        lto_download_error(msg);
        if (lto_pid > 0)
        {
            printf("lto_wait_for_stop()/%d kill(%d,SIGKILL %d)\n",
                                    __LINE__, lto_pid, SIGKILL);
            kill(lto_pid, SIGKILL);
        }
    }
    else
    {
        lto_download_error("LTO Download Done");
    }
    gdk_input_remove(lto_cb_tag);
    close(lto_fd);
    lto_fd = -1;
}

//-----------------------------------------------------------------------------
//
//              lto_download_stop()
//
//      Forcibly stop the LTO download.
//
//-----------------------------------------------------------------------------

void
lto_download_stop()
{
    if (lto_pid > 0)
    {
        printf("lto_download_stop()/%d kill(%d,SIGKILL %d)\n",
                                __LINE__, lto_pid, SIGKILL);
        kill(lto_pid, SIGKILL);
        lto_wait_for_stop();
    }
}

//-----------------------------------------------------------------------------
//
//                      ltoDownloadPidWait()
//
//      Wait for the LTO download PID to complete.
//
//      Returns:
//              0       Done.  Successfull.
//              msg     Some error message.
//
//-----------------------------------------------------------------------------

static const char*
ltoDownloadPidWait()
{
    int   stat = 0;
    pid_t x    = waitpid(lto_pid, &stat, 0);                // WNOHANG will poll
    if (x == lto_pid)
    {
        lto_downloading_flag = false;
        lto_pid = 0;
        lto_download_done();
    }
    else
    {
        fprintf(stderr, "omgui: %d = waitpid(%d) err %d %s\n",
                x, lto_pid, errno, strerror(errno));
    }
    if (WIFEXITED(stat))
    {
        lto_download_done();
        lto_downloading_flag = false;
        lto_pid = 0;

        int exit_code = WEXITSTATUS(stat);
        fprintf(stderr, "omgui: %s exit code %d\n",
                        cc_lto_get, exit_code);
        if (exit_code) return "lto_get failed: abnormal exit";
        return 0;
    }
    if (WIFSIGNALED(stat))
    {
        return "lto_get failed: signalled";
    }
    return "lto_get wait failed";
}
