/*
 * Copyright © 2007 Global Locate.
 *
 * Written by Ken Yale.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <stdlib.h>
#include <string.h>
#include "agps.h"
#include "sv_bar.h"

void notebook_expose(const char* pageName);

static GtkDrawingArea* AZ;
static GdkPixmap*      AZ_PIX = 0;
static PangoLayout*    AZ_text;

static GdkGC* blue_gc = 0;
static GdkGC* grey_gc = 0;

extern GdkGC* alloc_color(GdkPixmap* pm, const char* color);

enum
{
        MAX_AZ_SEGMENTS = 22
};

static GdkSegment az_lines[MAX_AZ_SEGMENTS];
static int        nAzLines = 0;

static void
AZ_add_line(int x0, int y0, int x1, int y1)
{
    if (nAzLines >= MAX_AZ_SEGMENTS)
    {
        // printf("az_add_line(%d,%d  %d,%d) discarded %d\n",
        //             x0, y0, x1, y1, nAzLines);
        return;
    }
    // printf("az_add_line(%d,%d  %d,%d) %d\n",
    //                 x0, y0, x1, y1, nAzLines);
    GdkSegment& s = az_lines[nAzLines++];
    s.x1 = x1;
    s.y1 = y1;
    s.x2 = x0;
    s.y2 = y0;
}

// #define DEBUG

#ifdef  DEBUG
static void
debugGC_(int line)
{
    printf("%d) blue_gc 0x%x  grey_gc 0x%x  AZ_PIX 0x%x\n",
                line, blue_gc, grey_gc, AZ_PIX);
    fflush(stdout);
}
#define debugGC()       debugGC_(__LINE__)
#else
#define debugGC()
#endif

bool showTrack = false;

static void
on_clicked_show_sv_track(GtkWidget* w, gpointer data)
{
    const char* text = "Show Track";
    if (showTrack)
    {
        text = "Disable Track";
    }
    gtk_button_set_label(GTK_BUTTON(w), text);
    showTrack = !showTrack;
}

static void
AZ_alloc_colors()
{
    if (!AZ_PIX) return;
    if (!blue_gc) blue_gc = alloc_color(AZ_PIX, "Blue");
    if (!grey_gc) grey_gc = alloc_color(AZ_PIX, "SlateGrey");
}

static void
AZ_clear()
{
    if (!AZ_PIX) return;
    GtkWidget* w = GTK_WIDGET(AZ);
    gdk_draw_rectangle(AZ_PIX, w->style->white_gc, TRUE,
                       0, 0,
                       w->allocation.width,
                       w->allocation.height);
}

static gboolean
configure_AZ_event(GtkWidget* w, GdkEventConfigure* event)
{
    debugGC();
    if (AZ_PIX)
    {
        g_object_unref(AZ_PIX);
    }

    //--------------------------------------------------------------------------
    //          pixmap
    //--------------------------------------------------------------------------
    AZ_PIX = gdk_pixmap_new(w->window,
                            w->allocation.width,
                            w->allocation.height, -1);
    AZ_clear();
    debugGC();
    return TRUE;
}

static gboolean
expose_AZ_event(GtkWidget* w, GdkEventExpose* event)
{
    notebook_expose("AZ");
    if (AZ_PIX)
    {
        GdkRectangle& a = event->area;
        gdk_draw_drawable(w->window,
                      w->style->fg_gc[GTK_WIDGET_STATE(w)],
                      AZ_PIX, a.x, a.y, a.x, a.y, a.width, a.height);
    }
    return FALSE;
}

extern GdkGC* page_SS_getGC(bool used);

static GdkGC*
getGC(bool used)
{
    AZ_alloc_colors();
    GdkGC*    gc = grey_gc;
    if (used) gc = blue_gc;
    if (!gc)  gc = GTK_WIDGET(AZ)->style->black_gc;
    return gc;
}

enum
{
    CENTER_XY = 3,
    CENTER_X  = 1,
    CENTER_Y  = 2,
    CENTER_NONE = 0,

    ALIGN_BOTTOM = 16,
    ALIGN_RIGHT  = 32
};

static int
AZ_draw_text(int x, int y, const char* t,
                int center = CENTER_XY, GdkGC* gc = 0)
{
    int w = 0;
    int h;

    if (!gc) gc = getGC(false);
    pango_layout_set_text(AZ_text, t, -1);
    pango_layout_get_pixel_size(AZ_text, &w, &h);

    if (center)
    {
        if      (center & CENTER_X    ) x -= w/2;
        else if (center & ALIGN_RIGHT ) x -= w;
        if      (center & CENTER_Y    ) y -= h/2;
        else if (center & ALIGN_BOTTOM) y -= h;
    }

    // printf("AZ_draw_text(%d,%d,\"%s\", %d)\n", x, y, t, center);
    gdk_draw_layout(AZ_PIX, gc, x, y, AZ_text);

    return w;
}

static void
AZ_draw_circle(int r, bool drawText = true)
{
    GdkGC* gc = getGC(false);

    int radius = SV_bar::scale_radius(r);
    int x = LARGE_WINDOW_WIDTH/2 - radius;
    int y = x;
    int width  = radius * 2;
    int height = width;
    int arc0 = 0;
    int arcN = 360 * DEGREE_64;
    gdk_draw_arc(AZ_PIX, gc, FALSE,
                x, y,                 // starting point
                width, height,
                arc0, arcN);

    if (drawText)
    {
        char elevation[4];
        sprintf(elevation, "%d", r);
        AZ_draw_text(x, LARGE_WINDOW_WIDTH/2, elevation);
    }
}

static void
AZ_draw_radial(int theta, int radius, bool drawText)
{
    SV_bar r;
    r.el = radius;
    r.az = theta;

    int x0;
    int y0;
    r.get_xy(x0, y0);

    y0  = LARGE_WINDOW_WIDTH/2 - y0;
    x0 += LARGE_WINDOW_WIDTH/2;

    int x1;
    int y1;
    r.el = 0;
    r.get_xy(x1, y1);

    y1  = LARGE_WINDOW_WIDTH/2 - y1;
    x1 += LARGE_WINDOW_WIDTH/2;

    // if (drawText)
    //    printf("radial(%d,%d) x1,y1 = {%d,%d}\n", theta, radius, x1, y1);

    if (drawText)
    {
        r.el = -8;
        r.get_xy(x1, y1);
        y1  = LARGE_WINDOW_WIDTH/2 - y1;
        x1 += LARGE_WINDOW_WIDTH/2;

        char str[8];
        sprintf(str, "%d", theta);
        AZ_draw_text(x1, y1, str);
    }
    else
    {
        AZ_add_line(x0, y0, x1, y1);
    }
}

static void
AZ_draw_sv(SV_bar& sv, bool drawDot)
{
    int yy;
    int xx;
    sv.get_xy(xx, yy);

    yy  = LARGE_WINDOW_WIDTH/2 - yy;
    xx += LARGE_WINDOW_WIDTH/2;

    char prn[4];
    sprintf(prn, "%d", sv.prn);

    GdkGC* gc = getGC(sv.use);

    // Get the text width.
    int w;
    int h;
    pango_layout_set_text(AZ_text, prn, -1);
    pango_layout_get_pixel_size(AZ_text, &w, &h);

    //----------------------------------------------
    //          Put a circle down first
    //
    //  SV used:    dark blue circle
    //  SV unused:  light grey circle
    //----------------------------------------------
    if (h > w) w = h;                   // w = max(w,h)
    int radius = 3 + w/2;
    if (drawDot)
    {
        gdk_draw_arc(AZ_PIX, gc, TRUE,
                    xx-radius, yy-radius-1,
                    radius*2, radius*2,
                    0, 360 * DEGREE_64);
    }
    else
    {
        //----------------------------------------------
        //          Draw the SV number on the dot
        //
        //  SV used:    white text
        //  SV unused:  black text
        //----------------------------------------------
        gc = GTK_WIDGET(AZ)->style->black_gc; 
        if (sv.use) gc = GTK_WIDGET(AZ)->style->white_gc; 
        AZ_draw_text(xx, yy, prn, CENTER_XY, gc);
    }

    // printf("AZ_draw_sv(%d   el %d    az %d) at %d,%d\n",
    //              sv.prn, sv.el, sv.az, xx, yy);
}

static void
AZ_draw_svs()
{
    if (!AZ_PIX) return;
    if (!showTrack)
    {
        AZ_clear();
    }

    int f = LARGE_WINDOW_WIDTH;
    int h = f/2;
    int fw = FONT_WIDTH;
    int fh = FONT_HEIGHT;

    static bool been_here = false;
    if (!been_here)
    {
        been_here = true;

        AZ_add_line(fw,h,  f-fw,h);            // horizontal line
        AZ_add_line(h,fh,  h,f-fh);            // vertical   line
        for (int theta = 15; theta < 360; theta += 15)
        {
            int  r    = 44;
            switch (theta)
            {
            default: break;
            case 90: case 180: case 270: continue;
            case 45: case 45 + 90: case 45 + 180: case 45 + 270:
                r    = 74;
                break;
            }
            AZ_draw_radial(theta, r, false);
        }
    }

    GdkGC* gc = getGC(false);

    gdk_draw_segments(AZ_PIX, gc, az_lines, nAzLines);
    AZ_draw_radial(45    , 74, true);
    AZ_draw_radial(45+ 90, 74, true);
    AZ_draw_radial(45+180, 74, true);
    AZ_draw_radial(45+270, 74, true);

    AZ_draw_circle(0, false);
    AZ_draw_circle(15);
    AZ_draw_circle(45);
    AZ_draw_circle(75);

    AZ_draw_text(h, 0, "N", CENTER_X,                gc);
    AZ_draw_text(h, f, "S", CENTER_X | ALIGN_BOTTOM, gc);
    AZ_draw_text(f, h, "E", CENTER_Y | ALIGN_RIGHT,  gc);
    AZ_draw_text(0, h, "W", CENTER_Y,                gc);

    // Draw all the dots.
    // Then draw all the numbers.
    for (int b = 0; b < SV_bar::nSS; ++b) AZ_draw_sv(bars[b], true);
    for (int b = 0; b < SV_bar::nSS; ++b) AZ_draw_sv(bars[b], false);

    gtk_widget_queue_draw_area(GTK_WIDGET(AZ), 0, 0,
                               PIXMAP_WIDTH, PIXMAP_WIDTH);
}

void
page_AZ_done()
{
    AZ_draw_svs();
}

GtkWidget*
page_AZ_create(GtkWidget* window)
{
    //--------------------------------------------------------------------------
    //          table
    //--------------------------------------------------------------------------
    GtkWidget* table = gtk_table_new(14, 2, TRUE);

    //--------------------------------------------------------------------------
    //          label
    //--------------------------------------------------------------------------
    GtkWidget* label = gtk_label_new("Satellite Position");
    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE (table), label, 0, 2, 0, 1);
    gtk_widget_show(label);

    AZ_text = gtk_widget_create_pango_layout(label, "");

    //--------------------------------------------------------------------------
    //          window for the satellite position
    //--------------------------------------------------------------------------
    GtkWidget* w  = gtk_drawing_area_new();
    AZ = GTK_DRAWING_AREA(w);
    gtk_drawing_area_size(AZ, PIXMAP_WIDTH, PIXMAP_WIDTH);

    gtk_signal_connect(GTK_OBJECT(w), "configure_event",
                    G_CALLBACK(configure_AZ_event), w);

    gtk_signal_connect(GTK_OBJECT(w), "expose_event",
                    G_CALLBACK(expose_AZ_event), w);

    gtk_widget_set_events(w, GDK_EXPOSURE_MASK);
    gtk_table_attach_defaults(GTK_TABLE (table), w, 0, 2, 1, 12);
    gtk_widget_show(w);

    //--------------------------------------------------------------------------
    //          Show SV track
    //--------------------------------------------------------------------------
    GtkWidget* button = gtk_button_new_with_label("Show Track");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_show_sv_track), button);
    gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 13, 14);
    gtk_widget_show(button);

    gtk_widget_show(table);
    return table;
}

void
page_AZ_report(FILE* f)
{
    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        SV_bar& sv = bars[b];
        fprintf(f, "PRN %2d  Elevation %2d   Azimuth %3d  %s\n",
                sv.prn, sv.el, sv.az, (sv.use) ? " " : " unused");
    }
}
