/*
 * Copyright © 2007 Global Locate.
 *
 * Written by Ken Yale.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <fcntl.h>
#include "agps.h"
#include "sv_bar.h"
#include "lto_download.h"
#include "liblto.h"             // Actual LTO library code from GL.

static GdkGC* blue_gc = 0;
static GdkGC* grey_gc = 0;

static GtkDrawingArea* LTO;
static GdkPixmap*      LTO_PIX = 0;
static PangoLayout*    LTO_text;

extern "C" void page_lto_file_change();

extern void notebook_expose(const char* pageName);

static void lto_set_download_text(const char* msg);
static void lto_redraw();
static void lto_pixmap_update();
static void lto_draw_percent(bool age, int percent);

GtkLabel* LTO_label;
GtkLabel* LTO_age_label;

int lto_age_percent = 0;
static char lto_download_progress_ = ' ';
static int  lto_download_percent_  =   0;
static void lto_file_lease(bool getLease);

static void
lto_report(FILE* f)
{
    char msg[128];
    int hours = 0;
    const char* ltoFile = "/home/root/gpsd/lto2.dat";
    const char* err = NULL; /*ltoGetExpiration( ltoFile, &hours,
                                        msg, sizeof(msg));*/
    const char* report = err;
    if (!report) report = msg;

    if (f) fprintf(f, "%s\n", report);
    else   gtk_label_set_text(LTO_age_label, report);

    printf("%s = ltoGetExpiration(%s, %d, %s)\n",
                (err) ? err : "OK", 
                ltoFile,
                hours,
                msg);

    if (err)
    {
        lto_age_percent = 0;
    }
    else
    {
        const int max_hours = 7 * 24;           // Assume 7-day LTO.
        int percent = hours * 100;

        percent += max_hours/2;
        percent /= max_hours;

        lto_age_percent = percent;
        printf("Percent %d\n", lto_age_percent);
    }

    ltoLastLoaded(sizeof(msg), msg);

    if (f) fprintf(f, "%s\nLTO useful for 7 days x %d%%", msg, lto_age_percent);
    else   gtk_label_set_text(LTO_label, msg);
}

//------------------------------------------------------------------------------
//
//      Update a label with the sad news
//
//------------------------------------------------------------------------------

void
lto_download_error(const char* msg)
{
    gtk_label_set_text(LTO_label, msg);
}

//------------------------------------------------------------------------------
//
//      Update the LTO progress message and the.
//      LTO percentage bar.
//
//------------------------------------------------------------------------------

static void
lto_draw_download_progress()
{
    char buf[64];
    sprintf(buf, "Download %d%% complete %c.",
                                lto_download_percent_, lto_download_progress_);
    lto_set_download_text(buf);
    lto_draw_percent(false, lto_download_percent_);
}

void
lto_download_progress(char progress)
{
    lto_download_progress_ = progress;
    lto_redraw();
}

void
lto_download_percent(int n)
{
    lto_download_percent_ = n;
    lto_redraw();
}

static void
on_clicked_factory_reset(GtkWidget* w, gpointer d)
{
    const char* msg = ltoRemove();
    if (!msg) msg = "Reset done";
    lto_download_error(msg);
    lto_report(0);
    lto_redraw();
}

static GdkGC* 
getGC(bool used)
{
    GdkGC* gc = grey_gc;
    if (used)
    {
        gc = blue_gc;                   // used SV is colored blue
    }
    if (!gc)
    {
        gc = GTK_WIDGET(LTO)->style->black_gc;
    }
    return gc;
}

static void
lto_set_download_text(const char* msg)
{
    int x = 20;
    int y = 85;

    pango_layout_set_text(LTO_text, msg, -1);
    gdk_draw_layout(LTO_PIX, getGC(false), x, y, LTO_text);

    // printf("LTO_set_download [%d,%d] %s\n", x, y, msg);
}

extern void lto_download_stop();
bool downloading = false;
GtkButton* lto_download_button = 0;

void
lto_download_done()
{
    downloading = false;
    gtk_button_set_label(lto_download_button, "Start Download");
    lto_report(0);
}

static void
on_clicked_start_download(GtkWidget* w, gpointer d)
{
    // printf("clicked download %d\n", downloading);
    if (downloading)
    {
        lto_download_stop();
        return;
    }

    lto_download_percent_ = 0;
    const char* msg = ltoDownloadStart();

    if (msg)
    {
        lto_set_download_text(msg);
        lto_pixmap_update();
    }
    else
    {
        lto_set_download_text("LTO download started");
        lto_report(0);
        downloading = true;

        gtk_button_set_label(lto_download_button, "Stop Download");

        lto_redraw();
    }
}

static void
lto_pixmap_clear()
{
    GtkWidget* w = GTK_WIDGET(LTO);
    gdk_draw_rectangle(LTO_PIX, w->style->white_gc, TRUE,
                       0, 0,
                       w->allocation.width,
                       w->allocation.height);
}

static gboolean
event_configure_lto(GtkWidget* w, GdkEventConfigure* event)
{
    if (LTO_PIX)
    {
        g_object_unref(LTO_PIX);
    }

    //--------------------------------------------------------------------------
    //          pixmap to make the strength bar meter easier to draw
    //--------------------------------------------------------------------------
    LTO_PIX = gdk_pixmap_new(w->window,
                            w->allocation.width,
                            w->allocation.height, -1);
    // printf("LTO pixmap %d x %d\n", w->allocation.width, w->allocation.height);
    lto_pixmap_clear();
    return TRUE;
}

static gboolean
event_expose_lto(GtkWidget* w, GdkEventExpose* event)
{
    notebook_expose("LT");

    //  TODO:  figure out why we're being called so frequently:

    lto_redraw();

    GdkRectangle& a = event->area;
    gdk_draw_drawable(w->window,
                      w->style->fg_gc[GTK_WIDGET_STATE(w)],
                      LTO_PIX, a.x, a.y, a.x, a.y, a.width, a.height);
    return FALSE;
}

static GdkGC*
alloc_color(GdkPixmap* pm, const char* color)
{
    GdkGC* result = 0;
    if (!pm) return result;

    GdkColormap* map = gdk_colormap_get_system();

    GdkColor blue;
    if (TRUE == gdk_color_parse(color, &blue))
    {
        if (TRUE == gdk_colormap_alloc_color(map, &blue, FALSE, TRUE))
        {
            GdkGCValues v;
            v.foreground = blue;
            result = gdk_gc_new_with_values(pm, &v, GDK_GC_FOREGROUND);
            // printf("GC for %s 0x%x\n", color, result);
        }
        else printf("Can't alloc %s\n", color);
    }
    else printf("Can't parse %s\n", color);
    return result;
}

static void
page_lto_alloc_colors()
{
    if (!LTO_PIX) return;
    if (!blue_gc) blue_gc = alloc_color(LTO_PIX, "Blue");
    if (!grey_gc) grey_gc = alloc_color(LTO_PIX, "SlateGrey");
}

static void
lto_draw_percent(bool age, int percent)
{
    page_lto_alloc_colors();
    GdkGC* gc = GTK_WIDGET(LTO)->style->black_gc;

    int line_width = 1;
    int adj        = 4;
    int x = 20;
    int y = 35;
    int h =  40;
    int w = 400;

    if (!age)
    {
        y += h * 2;             // draw the download progress
    }

    //----------------------------------------------
    //      Draw the rectangle around the percent
    //          (line width is implied in the GC)
    //----------------------------------------------

    gdk_draw_rectangle(LTO_PIX, gc, FALSE, x, y, w, h);

    if (percent > 100) percent = 100;
    if (percent <=  0) return;

    //----------------------------------------------
    //    Draw the percentage
    //----------------------------------------------

    x += adj;
    y += adj;
    h -= 2 * adj - line_width;
    w -= 2 * adj - line_width;

    int progress  = 50 + (w * percent);
        progress /= 100;
    gc = getGC(true);

    gdk_draw_rectangle(LTO_PIX, gc, TRUE, x, y, progress, h);

    //----------------------------------------------
    //    Draw the percentage text
    //
    // % < 60:  Draw to the right of the bar.
    // % > 60:  Draw to the left of the bar.
    //----------------------------------------------


    char buf[12];
    sprintf(buf, "%d%%", percent);
    pango_layout_set_text(LTO_text, buf, -1);

    int w_text;
    int h_text;
    pango_layout_get_pixel_size(LTO_text, &w_text, &h_text);

    y += (h-h_text)/2;
    if (percent < 60)
    {
        x += 4 + progress;
        // gc = getGC(true);            // blue on white looks OK.
    }
    else
    {
        x -= (4 - progress + w_text);
        gc = getGC(false);              // light grey on blue looks OK.
    }

    // printf("lto_draw_percent(%s,%d) \"%s\" at (%d %d)\n",
    //            (age) ? "age" : "download", percent, buf, x, y);
    gdk_draw_layout(LTO_PIX, gc, x, y, LTO_text);
}

static void
lto_pixmap_update()
{
    gtk_widget_queue_draw_area(GTK_WIDGET(LTO), 0, 0, PIXMAP_WIDTH, PIXMAP_WIDTH);
}

static void
lto_redraw()
{
    if (!LTO_PIX)
    {
        // printf("lto_redraw()/%d: no LTO_PIX\n", __LINE__);
        return;
    }
    lto_pixmap_clear();
    lto_draw_percent(true, lto_age_percent);
    if (downloading) lto_draw_download_progress();
    lto_pixmap_update();
}

void
page_LTO_init()
{
}

void
page_LTO_done()
{
    lto_redraw();
}

GtkWidget*
page_LTO_create(GtkWidget* window)
{
    GtkWidget* label;
    GtkWidget* button;
    GtkWidget* w;
    GtkWidget* table;

    //--------------------------------------------------------------------------
    //          table
    //--------------------------------------------------------------------------
    table = gtk_table_new(14, 2, TRUE);

    //--------------------------------------------------------------------------
    //          Page title label
    //--------------------------------------------------------------------------
    label = gtk_label_new("LTO Status");
    gtk_label_set_justify((GtkLabel*)label, GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE (table), label, 0, 2, 0, 1);
    gtk_widget_show(label);
    LTO_age_label = GTK_LABEL(label);

    //--------------------------------------------------------------------------
    //          label
    //--------------------------------------------------------------------------
    label = gtk_label_new("Expired");
    gtk_label_set_justify((GtkLabel*)label, GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE (table), label, 0, 2, 1, 2);
    gtk_widget_show(label);
    LTO_label = GTK_LABEL(label);
    LTO_text = gtk_widget_create_pango_layout(label, "");

    //--------------------------------------------------------------------------
    //          window for the LTO bar
    //--------------------------------------------------------------------------
    w  = gtk_drawing_area_new();
    LTO = GTK_DRAWING_AREA(w);
    gtk_drawing_area_size(LTO, 476, 200);
    gtk_signal_connect(GTK_OBJECT(w), "configure_event",
                    G_CALLBACK(event_configure_lto), w);
    gtk_signal_connect(GTK_OBJECT(w), "expose_event",
                    G_CALLBACK(event_expose_lto), w);
    gtk_widget_set_events(w, GDK_EXPOSURE_MASK);
    gtk_table_attach_defaults(GTK_TABLE (table), w, 0, 2, 3, 8);
    gtk_widget_show(w);

    //--------------------------------------------------------------------------
    //          Start Download button
    //--------------------------------------------------------------------------
    button = gtk_button_new_with_label("Start Download");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_start_download), button);
    gtk_table_attach_defaults(GTK_TABLE (table), button, 0, 1, 13, 14);
    gtk_widget_show(button);
    lto_download_button = GTK_BUTTON(button);

    //--------------------------------------------------------------------------
    //          Factory Reset button
    //--------------------------------------------------------------------------
    button = gtk_button_new_with_label("Factory Reset");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_factory_reset), button);
    gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 13, 14);
    gtk_widget_show(button);

    gtk_widget_show(table);

    lto_report(0);
    lto_file_lease(true);

    return table;
}

//-------------------------------------------------------------------------------
//
//      Experimental feature to update the screen if the LTO file changes.
//      Since the use of SIGIO inside GTK+ is unknown, I have a very
//      risky signal forwarding chaining/cascading/forwarding algorithm.
//      I can't find any references to this technique, so perhaps it won't
//      work.
//
//      If omgui starts acting strangely, then perhaps use of SIGIO is the
//      problem, and we can't forward it in this manner.  Fallbacks are:
//
//      0)  Drop this entire feature.  It is uncertain what to do with the
//          output anyway.
//      1)  Use a different signal.  Scan thru all the SIGUSRn until
//          we find one that doesn't have a handler.
//      2)  Fix this forwarding feature.
//
//-------------------------------------------------------------------------------

#define OPT_LTO_LEASE
#ifdef OPT_LTO_LEASE                    // {

static void lto_file_lease(bool getLease);
extern const char* cc_file;                     // See ltolib.cpp

#if 0
static bool
sig_function(void* f)
{
    if ((void*)SIG_IGN == f) return false;
    if ((void*)SIG_DFL == f) return false;
    return true;
}

static void
sig_lto_file_changed(int sigNum, struct siginfo* info, void* v)
{
    // Undocumented signal feature:
    // ucontext_t* u = (ucontext_t*) v;

    if ((info->si_fd == lto_file_fd) && (-1 != lto_file_fd))
    {
        printf("LTO file(%d) %s changed\n", lto_file_fd, cc_file);

        // Free the lease right away, and then schedule an update
        // of the screen info for the near future (somehow)...
        lto_file_changed = true;
        lto_file_lease(false);
        return;
    }

    // Not our signal, forward it...
    if (old_io.sa_flags & SA_SIGINFO)     // new signal style handler
    {
        if (sig_function((void*)old_io.sa_sigaction))
        {
            old_io.sa_sigaction(sigNum, info, v);
        }
        // else  not a function, so don't call
    }
    else                    // old style handler
    {
        if (sig_function((void*)old_io.sa_handler))
        {
            old_io.sa_handler(sigNum);
        }
    }
}
#endif

#define  OPT_LTO_LEASE
#undef   OPT_LTO_LEASE

static void
lto_file_lease(bool getLease)
{
#ifdef  OPT_LTO_LEASE
    const char* lto_file = cc_file;

    printf("lto_file_lease(%d) %s have lease\n", getLease,
                (-1 == lto_file_fd) ? "do not" : "DO");

    if (getLease)
    {
        if (-1 != lto_file_fd) return;          // already have the lease
        lto_file_fd = open(lto_file, O_RDONLY);

        lto_file_changed = false;

        fcntl(lto_file_fd, F_SETLEASE, F_RDLCK);

        struct sigaction new_io;
        new_io.sa_sigaction = sig_lto_file_changed;
        new_io.sa_flags   = SA_SIGINFO;
        sigemptyset(&new_io.sa_mask);
        sigaction(SIGIO, &new_io, &old_io);
    }
    else                // free the lease
    {
        if (-1 == lto_file_fd) return;          // don't have the lease.
        fcntl(lto_file_fd, F_SETLEASE, F_UNLCK);
        close(lto_file_fd);
        sigaction(SIGIO, &old_io, 0);
        lto_file_fd = -1;
    }
#else
    printf("lto_file_lease(%d)  OPT_LEASE_LTO undefined\n", getLease);
#endif
}

#else                                   // } {

static void
lto_file_lease(bool getLease)
{
    printf("NOT SUPPORTED: lto_file_lease(%d) \n", getLease);
}

#endif                                  // }



void
page_LTO_report(FILE* f)
{
    lto_report(f);
}
