/*
 * Copyright © 2007 Global Locate.
 *
 * Written by Ken Yale.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "agps.h"
#include "sv_bar.h"

SV_bar bars[MAX_SVS];
int SV_bar::nSS = 0;
bool SV_bar::changed = false;
int SV_bar::leastSamplesAveraged = 0;

static GdkGC* blue_gc = 0;
static GdkGC* grey_gc = 0;

static GtkDrawingArea* SS;
static GdkPixmap*      SS_PIX = 0;
int dbm = 0;
static PangoLayout* SS_text;
GtkWidget* top4_button;

//      KLUDGE for A3 or larger fonts only:
// bool large_font  = true;
// bool a3_hardware = true;
#define large_font  1
#define a3_hardware 1

inline int
format_ss(int dbHz)
{
    if (dbm) dbHz += DB_HZ_to_DBM;

    if (large_font || a3_hardware)
    {
        if (dbHz < 0) dbHz = -dbHz;
    }
    return dbHz;
}

inline double
format_ssf(double dbHz)
{
    if (dbm) return dbHz + DB_HZ_to_DBM;
    return dbHz;
}

extern void notebook_expose(const char* pageName);

static void SS_draw_bars();
void page_SS_use_prn(int prn);
void page_SS_have_prn(int prn, int ss, int el, int az);
static double get_top_four_SVS();
unsigned long msTime(void);

//  Return a number between -100 and +100 along the X and Y axes.

const double PI = 3.1415926535;

int SV_bar::scale_radius(int r)
{
    r = 90 - r;
    r *= LARGE_WINDOW_WIDTH/2;
    r += 45;
    return r/90;
}

void SV_bar::get_xy(int&x, int& y)
{
    float radius = (float) scale_radius(el);
    float theta  = (float) -(az - 90);
    float rad    = theta * PI/180.0;

    float fx = radius * cos(rad);
    float fy = radius * sin(rad);
    x = (int) (fx + 0.5);
    y = (int) (fy + 0.5);

    //printf("get_xy:  r %f,  theta %f   rad %f   fx %f   fy %f\n",
    //                 radius,theta,     rad,     fx,     fy);
}

GtkLabel* ss_label;

void SS_reset_max_min()
{
    unsigned long ms = msTime();
    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        SV_bar& sv = bars[b];
        if (!sv.use) continue;

        sv.ss_max_time = ms;
        sv.ss_min_time = ms;
        sv.ss_max      = sv.ss;
        sv.ss_min      = sv.ss;
    }
}

void SS_reset_average()
{
    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        SV_bar&  sv = bars[b];
        sv.nAvg     = 0;
        sv.sumAvg   = 0;
        sv.indexAvg = 0;
    }
}

static void
on_clicked_reset_max_min(GtkWidget* w, gpointer d)
{
    SS_reset_max_min();
    SS_draw_bars();
}

static void
reset_top4()
{
    gtk_button_set_label(GTK_BUTTON(top4_button), "Top 4 SV");
}

static void
on_clicked_top4(GtkWidget* w, gpointer data)
{
    double top4     = format_ssf(get_top_four_SVS());
    int    nSamples = SV_bar::leastSamplesAveraged;
    if (nSamples)
    {
        char buf[64];
        sprintf(buf,"Top 4 %.1f (%d)", top4, nSamples);
        gtk_button_set_label(GTK_BUTTON(w), buf);
    }
    else
    {
        reset_top4();
    }
}

static void
on_clicked_reset_average(GtkWidget* w, gpointer d)
{
    SS_reset_average();
    reset_top4();
    SS_draw_bars();
}

static void
on_ss_units_change()
{
    const char* units = "Signal Strength (dB.Hz)";
    if (dbm) units = "Signal Strength (dBm)";
    gtk_label_set_text(ss_label, units);
}

static void
on_clicked_dbm(GtkWidget* widget, gpointer data)
{
    dbm = 1 - dbm;
    gtk_button_set_label(GTK_BUTTON(data), (dbm) ? "dBm" : "C/N0");
    on_ss_units_change();
    reset_top4();
    SS_draw_bars();
}

static void
SS_clear()
{
    GtkWidget* w = GTK_WIDGET(SS);
    gdk_draw_rectangle(SS_PIX, w->style->white_gc, TRUE,
                       0, 0,
                       w->allocation.width,
                       w->allocation.height);
}

static gboolean
configure_SS_event(GtkWidget* w, GdkEventConfigure* event)
{
    if (SS_PIX)
    {
        g_object_unref(SS_PIX);
    }

    //--------------------------------------------------------------------------
    //          pixmap to make the strength bar meter easier to draw
    //--------------------------------------------------------------------------
    SS_PIX = gdk_pixmap_new(w->window,
                            w->allocation.width,
                            w->allocation.height, -1);
    SS_clear();
    return TRUE;
}

static gboolean
expose_SS_event(GtkWidget* w, GdkEventExpose* event)
{
    notebook_expose("SS");
    GdkRectangle& a = event->area;
    gdk_draw_drawable(w->window,
                      w->style->fg_gc[GTK_WIDGET_STATE(w)],
                      SS_PIX, a.x, a.y, a.x, a.y, a.width, a.height);
    return FALSE;
}

void
page_SS_done()
{
    if (!SV_bar::changed) return;
    SV_bar::changed = false;
    SS_draw_bars();
}

static int
SS_width(int maxWidth)
{
    int n = SV_bar::nSS;
    if (SV_bar::nSS < 1)
    {
        n = 1;
    }
    int seps = n+1;
    return (maxWidth - seps * SEP_WIDTH) / n;
}

static GdkGC* 
getGC(bool used)
{
    GdkGC* gc = grey_gc;
    if (used)
    {
        gc = blue_gc;                   // used SV is colored blue
    }
    if (!gc)
    {
        gc = GTK_WIDGET(SS)->style->black_gc;
    }
    return gc;
}

#if 0
static void
SS_draw_line(int x0, int y0, int x1, int y1)
{
    printf("AZ_draw_line(%d,%d,%d,%d)\n", x0, y0, x1, y1);
    gdk_draw_line(SS_PIX, getGC(false), x0, y0, x1, y1);
}
#endif

//------------------------------------------------------------------------------------
//
//      dbHz_to_pixels()
//
//      dBm     C/N0    Pixels  Use
//      -120      53    BAR_HEIGHT        Maximum
//      -133      40    <scaled properly> Typical
//      -163      10    0                 Minimum
//
//------------------------------------------------------------------------------------

int dbHz_to_pixels(int dbHz)
{
    if (dbHz >= 53) return BAR_HEIGHT;
    if (dbHz <= 10) return 0;
    dbHz -= 10;
    dbHz *= BAR_HEIGHT;
    dbHz += 21;                 // Rounding.
    dbHz /= 43;
    return dbHz;
}

GdkGC*
alloc_color(GdkPixmap* pm, const char* color)
{
    GdkGC* result = 0;
    if (!pm) return result;

    GdkColormap* map = gdk_colormap_get_system();

    GdkColor blue;
    if (TRUE == gdk_color_parse(color, &blue))
    {
        if (TRUE == gdk_colormap_alloc_color(map, &blue, FALSE, TRUE))
        {
            GdkGCValues v;
            v.foreground = blue;
            result = gdk_gc_new_with_values(pm, &v, GDK_GC_FOREGROUND);
            // printf("GC for %s 0x%x\n", color, result);
        }
        else printf("Can't alloc %s\n", color);
    }
    else printf("Can't parse %s\n", color);
    return result;
}

void
page_SS_alloc_colors()
{
    if (!SS_PIX) return;
    if (!blue_gc) blue_gc = alloc_color(SS_PIX, "Blue");
    if (!grey_gc) grey_gc = alloc_color(SS_PIX, "SlateGrey");
}

//-------------------------------------------------------------------------------------
//
//              SS_draw_el()
//
//  Draw a black filled arc showing the elevation at the bottom of the bar.
//  (x,y) should be the lower-left corner of the signal strength bar.
//
//  Max width is 40.
//
//-------------------------------------------------------------------------------------

static void
SS_draw_el(SV_bar& sv, int x, int y, int width)
{
    if (!sv.use) return;

    int arcN = sv.el * DEGREE_64;
    if (sv.el < MIN_ELEVATION) arcN = MIN_ELEVATION * DEGREE_64;

    const int max_width = 50;
    if (width > max_width)
    {
        x += (width - max_width)/2;
        width = max_width;
    }

    gdk_draw_arc(SS_PIX, GTK_WIDGET(SS)->style->black_gc, TRUE,
                x-width,  y-width,
                  width*2,  width*2,
                0, arcN);
}

enum
{
    SS_SMALL = 1,
    SS_TEXT  = 2
};

static void
SS_draw_ss(int ss, int y, int w, int flags)
{
    page_SS_alloc_colors();
    GdkGC* gc = getGC(false);

    int font_height = 16;               // TODO: see how to get this.
    int barHeightPixels = dbHz_to_pixels(ss);

    int base_y = y + BAR_HEIGHT + 2 * font_height;

    int h = 3;
    int r_y    = base_y - barHeightPixels - 1;

    //----------------------------------------------
    //          Draw the signal strength rule
    //----------------------------------------------
    if (flags & SS_SMALL)
    {
        ++r_y;
        h = 1;
    }
    gdk_draw_rectangle(SS_PIX, gc, TRUE, 0, r_y, w, h);

    //----------------------------------------------
    //          Draw the signal strength text
    //----------------------------------------------
    if (flags & SS_TEXT)
    {
        char dbmString[8];
        sprintf(dbmString, "%d", format_ss(ss));
        pango_layout_set_text(SS_text, dbmString, -1);
        if (flags & SS_SMALL) w += SEP_WIDTH;
        gdk_draw_layout(SS_PIX, getGC(false),
                        w + SEP_WIDTH, r_y - FONT_HEIGHT/2, SS_text);
    }
}

static void
SS_draw_bar(SV_bar& sv, int x, int y, int width, bool barOnly)
{
    int font_height = 16;               // TODO: see how to get this.

    int barHeightPixels = dbHz_to_pixels(sv.ss);

    page_SS_alloc_colors();

    //----------------------------------------------
    //          Draw the signal strength bar
    //----------------------------------------------

    int base_y = y + BAR_HEIGHT + 2*font_height;
    int r_y = base_y - barHeightPixels;
    int r_height = barHeightPixels;

    GdkGC* gc = getGC(sv.use);

    if (barOnly)
        gdk_draw_rectangle(SS_PIX, gc, TRUE, x, r_y, width, r_height);

    //----------------------------------------------
    //          Draw the elevation arc
    //----------------------------------------------

    SS_draw_el(sv, x, base_y, width);

    //----------------------------------------------
    //          Draw the minimum level (if it exists)
    //          Draw the maximim level (if it exists)
    //----------------------------------------------
    if (sv.use && barOnly)
    {
        GdkGC* black_gc = GTK_WIDGET(SS)->style->black_gc;
        if (sv.ss_min_time)
        {
            barHeightPixels = dbHz_to_pixels(sv.ss_min);
            r_y = base_y - barHeightPixels;
            gdk_draw_rectangle(SS_PIX, black_gc, TRUE,
                                x + width/4, r_y - 1, width/2, 3);
        }
        if (sv.ss_max_time)
        {
            barHeightPixels = dbHz_to_pixels(sv.ss_max);
            r_y = base_y - barHeightPixels;
            gdk_draw_rectangle(SS_PIX, black_gc, TRUE,
                                x, r_y - 1, width, 3);
        }
    }

    //----------------------------------------------
    //          Draw the average SS text
    //----------------------------------------------

    if (!barOnly)
    {
        int yy = y;
        int xx = x + 3;

        //----------------------------------------------
        //          Draw the signal strength average
        //          Draw the signal strength text
        //----------------------------------------------

        if (sv.use)
        {
            char dbmString[8];

            int i_dbm = sv.avg_ss();
            if (!i_dbm)
            {
                sprintf(dbmString, "-");
                xx += 4;
            }
            else
            {
                sprintf(dbmString, "%d", format_ss(sv.avg_ss()));
            }

            pango_layout_set_text(SS_text, dbmString, -1);
            gdk_draw_layout(SS_PIX, gc, xx, yy, SS_text);

            sprintf(dbmString, "%d", format_ss(sv.ss));
            pango_layout_set_text(SS_text, dbmString, -1);
            gdk_draw_layout(SS_PIX, gc, x+3, y + font_height + 4, SS_text);
        }

        //----------------------------------------------
        //          Draw the PRN text
        //----------------------------------------------

        char prn[4];
        sprintf(prn, "%d", sv.prn);
        pango_layout_set_text(SS_text, prn, -1);
        xx = (sv.prn > 9) ?
                    x + width/2 - 12                   // TODO: center this.
                :   x + width/2 -  5;                  // TODO: center this.
        yy = y + BAR_HEIGHT + 2*font_height;
        gdk_draw_layout(SS_PIX, gc, xx, yy, SS_text);
        // printf("PRN \"%s\" at {x %d  y %d}\n", prn, xx, yy);
    }
}

static void
SS_draw_bars()
{
    if (!SS_PIX) return;
    SS_clear();

    //  Leave some space to the right for the markers, etc.
    int w = SS_width(PIXMAP_WIDTH - 60);
    int indicator_w = SV_bar::nSS * (SEP_WIDTH + w) + SEP_WIDTH;
    int y = 10;

    //----------------------------------------------
    //          Draw the small SS horizontal lines
    //----------------------------------------------
    if (dbm)
    {
        SS_draw_ss(48, y, indicator_w, SS_SMALL);        // -125 dBm MAX
        SS_draw_ss(43, y, indicator_w, SS_TEXT);         // -130 dBm
        SS_draw_ss(38, y, indicator_w, SS_SMALL);
        SS_draw_ss(33, y, indicator_w, SS_TEXT);         // -140 dBm
        SS_draw_ss(28, y, indicator_w, SS_SMALL);
        SS_draw_ss(23, y, indicator_w, SS_TEXT);         // -150 dBm
        SS_draw_ss(18, y, indicator_w, SS_SMALL);
        SS_draw_ss(13, y, indicator_w, SS_TEXT);         // -160 dBm MIN
    }
    else
    {
        SS_draw_ss(50, y, indicator_w, SS_TEXT);        // 50 dBm MAX
        SS_draw_ss(45, y, indicator_w, SS_SMALL);
        SS_draw_ss(40, y, indicator_w, SS_TEXT);
        SS_draw_ss(35, y, indicator_w, SS_SMALL);
        SS_draw_ss(30, y, indicator_w, SS_TEXT);
        SS_draw_ss(25, y, indicator_w, SS_SMALL);
        SS_draw_ss(20, y, indicator_w, SS_TEXT);
        SS_draw_ss(15, y, indicator_w, SS_SMALL);
        SS_draw_ss(13, y, indicator_w, SS_SMALL | SS_TEXT);  // -160 dBm MIN
    }

    int x = SEP_WIDTH;
    if (SV_bar::nSS)
    {
        for (int b = 0; b < SV_bar::nSS; ++b)
        {
            SS_draw_bar(bars[b], x, y, w, true );
            x += SEP_WIDTH + w;
        }
    }

    x = SEP_WIDTH;
    if (SV_bar::nSS)
    {
        for (int b = 0; b < SV_bar::nSS; ++b)
        {
            SS_draw_bar(bars[b], x, y, w, false);
            x += SEP_WIDTH + w;
        }
    }

    //----------------------------------------------
    //          Draw the "avg" SS text
    //----------------------------------------------
    pango_layout_set_text(SS_text, "avg", -1);
    gdk_draw_layout(SS_PIX, GTK_WIDGET(SS)->style->black_gc, indicator_w, y, SS_text);

    // gtk_widget_queue_draw_area(GTK_WIDGET(SS), 0, 0, 400, 500);
    gtk_widget_queue_draw_area(GTK_WIDGET(SS), 0, 0, PIXMAP_WIDTH, PIXMAP_WIDTH);
}

//------------------------------------------------------------------------------
//
//         get_top_SV
//
//   Return the signal strength of the strongest SV
//   that is not in MASK
//
//------------------------------------------------------------------------------

#define BIT(b)  (1 << (b))

static int get_top_SV(int& mask)
{
    int our_mask = 0;
    int max = 0;
    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        int m = BIT(b);
        if (m & mask) continue;
        SV_bar& sv = bars[b];
        int avg = sv.avg_ss();
        if (avg > max)
        {
            our_mask = m;
            max = avg;
            int n = sv.nAvg;
            if (n > N_AVG) n = N_AVG;
            if (n < SV_bar::leastSamplesAveraged)
            {
                SV_bar::leastSamplesAveraged = n;
            }
        }
    }
    mask |= our_mask;
    printf("top 4:  0x%x %d\n", our_mask, max);
    return max;
}

static double get_top_four_SVS()
{
    int mask = 0;
    double sum = 0;
    int avg;
    SV_bar::leastSamplesAveraged = N_AVG;

    avg  = get_top_SV(mask);
    sum += avg;

    if (!avg)
    {
        SV_bar::leastSamplesAveraged = 0;
        return 0.0;
    }
    avg = get_top_SV(mask); sum += avg; if (!avg) return sum;
    avg = get_top_SV(mask); sum += avg; if (!avg) return sum/2.0;
    avg = get_top_SV(mask); sum += avg; if (!avg) return sum/3.0;

    return sum / 4.0;
}

static int
update_prn(int prn, int ss, int el, int az)
{
    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        if (prn == bars[b].prn)
        {
            SV_bar& sv = bars[b];
            if ((sv.el == el) && (sv.az == az) && (sv.ss == ss))
            {
                return 0;
            }
            if ((sv.el != el) || (sv.az != az))
            {
                // printf("Change PRN %d  el %d-->%d   az %d-->%d\n", 
                //             sv.prn, sv.el, el, sv.az, az);
                sv.el = el;
                sv.az = az;
            }
            sv.ss         = ss;
            sv.lastUpdate = msTime();
            if ((ss > sv.ss_max)) //  || ((sv.lastUpdate - sv.ss_max_time) > 1000*60))
            {
                sv.ss_max_time = sv.lastUpdate;
                sv.ss_max = ss;
            }

            //  Either we have a new minimum signal strength; OR
            //  We time-out the min signal
            if ((ss < sv.ss_min)) //  || ((sv.lastUpdate - sv.ss_min_time) > 1000*60))
            {
                sv.ss_min_time = sv.lastUpdate;
                sv.ss_min = ss;
            }
            return 1;
        }
    }
    return -1;
}

static int
findUnusedSV()
{
    unsigned long oldest_ms = msTime();
    int oldest = -1;

    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        if (bars[b].use) continue;
        if (bars[b].lastUpdate < oldest_ms)
        {
            oldest_ms = bars[b].lastUpdate;
            oldest    = b;
        }
    }
    return oldest;
}

unsigned long msTime(void);

static void
use_prn(int prn)
{
    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        SV_bar& sv = bars[b];
        if (prn == sv.prn)
        {
            if (!sv.use)
            {
                SV_bar::changed  = true;
                sv.use = true;
            }
            sv.add_ss();
            sv.lastUpdate = msTime();
            return;
        }
    }
}

void
page_SS_unuse_prns()
{
    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        bars[b].use = false;
    }
}

void
page_SS_use_prn(int prn)
{
    // printf("page_SS_use_prn(%d)\n", prn);
    if (prn <  1) return;
    if (prn > 32) return;               // Haven't launched PRN 33 yet.
    use_prn(prn);
}

void
page_SS_have_prn(int prn, int ss, int el, int az)
{
    // printf("page_SS_have_prn(%d %d   %d:%d) %d\n", prn, ss, el, az, SV_bar::nSS);
    if (prn <  1) return;
    if (prn > 32) return;               // Haven't launched PRN 33 yet.

    switch (update_prn(prn, ss, el, az))
    {
    case 0:
        return;
    case 1:
        SV_bar::changed = true;
        return;
    default:
        break;
    }

    int nextSV = SV_bar::nSS;
    if (SV_bar::nSS < MAX_SVS)
    {
        // TODO:  remove unused SVs.
        ++SV_bar::nSS;
    }
    else
    {
        nextSV = findUnusedSV();
        if (nextSV < 0) return;
    }

    // printf("add %d = {%d %d}\n", nextSV, prn, ss);
    SV_bar& sv = bars[nextSV];
    sv.init();
    sv.prn = prn;
    sv.ss  = ss;
    sv.el  = el;
    sv.az  = az;
    sv.lastUpdate = msTime();

    SV_bar::changed = true;
}

void
page_SS_init()
{
    memset(bars, 0, sizeof(bars));
    SV_bar::nSS = 0;
}

GtkWidget*
page_SS_create(GtkWidget* window)
{
    GtkWidget* label;
    GtkWidget* button;
    GtkWidget* w;
    GtkWidget* table;

    //--------------------------------------------------------------------------
    //          table
    //--------------------------------------------------------------------------
    table = gtk_table_new(14, 2, TRUE);

    //--------------------------------------------------------------------------
    //          label
    //--------------------------------------------------------------------------
    label = gtk_label_new("Signal Strength (db.Hz)");
    gtk_label_set_justify((GtkLabel*)label, GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE (table), label, 0, 2, 0, 1);
    gtk_widget_show(label);
    ss_label = GTK_LABEL(label);

    SS_text = gtk_widget_create_pango_layout(label, "");

    //--------------------------------------------------------------------------
    //          window for the strength bar meters
    //--------------------------------------------------------------------------
    w  = gtk_drawing_area_new();
    SS = GTK_DRAWING_AREA(w);
    gtk_drawing_area_size(SS, 500, 500);

    gtk_signal_connect(GTK_OBJECT(w), "configure_event",
                    G_CALLBACK(configure_SS_event), w);

    gtk_signal_connect(GTK_OBJECT(w), "expose_event",
                    G_CALLBACK(expose_SS_event), w);

    gtk_widget_set_events(w, GDK_EXPOSURE_MASK);
    gtk_table_attach_defaults(GTK_TABLE (table), w, 0, 2, 1, 12);
    gtk_widget_show(w);

#if 0
    //--------------------------------------------------------------------------
    //          Reset Average button
    //--------------------------------------------------------------------------
    button = gtk_button_new_with_label("Reset Max");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_reset_max_min), button);
    gtk_table_attach_defaults(GTK_TABLE (table), button, 0, 1, 12, 13);
    gtk_widget_show(button);
#else
    //--------------------------------------------------------------------------
    //          Reset max/min button
    //--------------------------------------------------------------------------
    GtkWidget* box = gtk_hbox_new (FALSE, 2);

    button = gtk_button_new_with_label("Reset max");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_reset_max_min), button);
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX (box), button, TRUE, TRUE, 1);

    //--------------------------------------------------------------------------
    //          Reset average button
    //--------------------------------------------------------------------------

    button = gtk_button_new_with_label("Reset avg");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_reset_average), button);
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX (box), button, TRUE, TRUE, 1);

    gtk_widget_show(box);
    gtk_table_attach_defaults(GTK_TABLE (table), box, 0, 1, 12, 13);
#endif

    //--------------------------------------------------------------------------
    //          dBm or C/N0 button
    //--------------------------------------------------------------------------
    button = gtk_button_new_with_label("dBm");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_dbm), button);
    gtk_table_attach_defaults(GTK_TABLE (table), button, 0, 1, 13, 14);
    gtk_widget_show(button);

    //--------------------------------------------------------------------------
    //          Top 4 SV Average button
    //--------------------------------------------------------------------------
    top4_button = gtk_button_new_with_label("");
    g_signal_connect(G_OBJECT(top4_button), "clicked",
                     G_CALLBACK(on_clicked_top4), top4_button);
    gtk_table_attach_defaults(GTK_TABLE(table), top4_button, 1, 2, 13, 14);
    reset_top4();
    gtk_widget_show(top4_button);

    gtk_widget_show(table);
    return table;
}

static void
SS_report_bar(SV_bar& sv, FILE* f)
{
    fprintf(f, " %2d", sv.prn);
    fprintf(f, "     %2d", sv.ss);

    int i_dbm = sv.avg_ss();
    if (i_dbm) fprintf(f, "   %2d", i_dbm);
    else       fprintf(f, "     ");

    if (sv.ss_min_time) fprintf(f, "   %2d", sv.ss_min);
    else                fprintf(f, "     ");

    if (sv.ss_max_time) fprintf(f, "   %2d", sv.ss_max);
    else                fprintf(f, "     ");

    fprintf(f, "         %2d", sv.el);
    fprintf(f, "      %3d",    sv.az);

    if (sv.use) fprintf(f, "  y\n");
    else        fprintf(f, "  UNUSED\n");
}

void
page_SS_report(FILE* f)
{
    fprintf(f, "PRN  dB.Hz  AVG  min  MAX  Elevation  Azimuth  Used\n");
    for (int b = 0; b < SV_bar::nSS; ++b)
    {
        SS_report_bar(bars[b], f);
    }
    fprintf(f, "\n");

    double top4     = format_ssf(get_top_four_SVS());
    int    nSamples = SV_bar::leastSamplesAveraged;
    if (nSamples)
    {
        fprintf(f, "Top 4 SVs %.1f (%d)\n", top4, nSamples);
    }
}
