/*
 * Copyright © 2007 Global Locate.
 *
 * Written by Ken Yale.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <gtk/gtk.h>
#include <stdlib.h>
#include "agps.h"

extern void notebook_expose(const char* pageName);

#ifndef AGPS_H_
#include "agps.h"
#endif 

extern void page_stress_on(void);
extern void page_stress_msg(const char* msg);

extern void create_range_controls(GtkWidget* window, GtkWidget* vbox, bool test);
extern const char* gllin_detect(void);
extern const char* gllin_start(void);
extern const char* gllin_stop(void);
extern void gllin_reset(char c);
extern void agps_cb_start(GtkWidget* label);
extern void agps_cb_stop(void);
extern void agps_stop(void);
extern void clearStats(void);
extern void page_pm_batch_start();

extern GtkLabel* agpsLabel;
extern void NMEA_MSG(const char* msg);

static bool gllinRunning = false;
bool doing_ptf = false;
extern bool doing_batch;
extern bool doing_stress;

//-------------------------------------------------------------------------------------
//
//              OMGUI versions
//
//-------------------------------------------------------------------------------------

#define OMGUI_MAJOR     0
#define OMGUI_MINOR     20

const char*
page_main_omgui_version(void)
{
    static char buf[64];
    sprintf(buf, "OpenMoko GPS UI v%d.%02d", OMGUI_MAJOR, OMGUI_MINOR);
    return buf;
}

//-------------------------------------------------------------------------------------
//
//              millisecond timer()
//
//-------------------------------------------------------------------------------------

unsigned long msTime(void)
{
    struct timespec t;
    clock_gettime(CLOCK_MONOTONIC, &t);
    unsigned long ms = t.tv_nsec/1000000;
    ms += t.tv_sec * 1000;
    return ms;
}

static unsigned long msStart;

void runtime_init()
{
    msStart = msTime();
}

unsigned long int runtime_ms()
{
    return msTime() - msStart;
}

unsigned long int runtime()
{
    return (runtime_ms() + 499) / 1000;
}

char testStartTime[16];

GtkWidget* startWidget = 0;
GtkWidget* ptfWidget   = 0;

char test_name[64];
extern void page_stress_set_report_name(void);

const char* page_main_test_name()
{
    if (test_name[0]) return test_name;
    return 0;
}

static char*
page_main_format_date(char* date, bool startTime)
{
    time_t aclock;
    time(&aclock);                      /* Get time in seconds */
    struct tm* t = localtime(&aclock);  /* Convert time to struct tm form */

    sprintf(date, "%02d:%02d %04d/%02d/%02d",
                    t->tm_hour, t->tm_min,
                    t->tm_year + 1900, t->tm_mon + 1, t->tm_mday);

    if (startTime)
    {
        sprintf(testStartTime, "%02d:%02d", t->tm_hour, t->tm_min);
        sprintf(test_name,     "%02d%02d",  t->tm_hour, t->tm_min);
        page_stress_set_report_name();
    }

    return date;
}

static char previousTestTime[32];

const char* nmea_epoch();

void
page_main_stop(const char* err)
{
    const char* msg = "";
    if (err)
    {
        msg = "error: ";
        NMEA_MSG(err);
    }
    else
    {
        err = "successfully";
    }
    agps_cb_stop();

    gtk_button_set_label(GTK_BUTTON(startWidget), "Power on");
    gllinRunning = false;
    gtk_widget_set_sensitive(ptfWidget, true);

    char date[32];
    page_main_format_date(date, false);

    char buf[2048];
    sprintf(buf, "Stopped gllin %s%s\n"
                 "at %s\n"
                 "run time %d seconds\n"
                 "Previous test data\n%s\n",
                     msg, err, date, (int) runtime(),
                     nmea_epoch());
    NMEA_MSG(buf);
}

//------------------------------------------------------------------------------
//
//              on_clicked_start()
//
//     Start or stop the GLLIN.
//
//------------------------------------------------------------------------------

extern void page_stress_set_report_name(void);
extern void page_pm_batch_done();

void on_clicked_start(GtkWidget* widget, gpointer data)
{
    widget = startWidget;
    const char* err;

    char buf[128];


    g_print("START button. GLLIN %srunning\n", (gllinRunning) ? "" : "not ");
    testStartTime[0] = 0;

    if (gllinRunning)
    {
        gllin_stop();
        page_pm_batch_done();
        return;
    }

    err = gllin_detect();
    if (err)
    {
        g_print("gllin_detect %s\n", err);
        NMEA_MSG(err);
        return;
    }

    doing_ptf = GPOINTER_TO_INT(data);
    err = gllin_start();
    if (err)
    {
        NMEA_MSG(err);
        return;
    }

    gllinRunning = true;

    clearStats();
    page_stress_set_report_name();
    page_main_format_date(previousTestTime, true);
    sprintf(buf, "Started GPS%s%s%s\nat %s\n",
                (doing_ptf)    ? " one fix"  : " periodic",
                (doing_batch)  ? " (batch)"  : "",
                (doing_stress) ? " (stress)" : "",
                previousTestTime);
    NMEA_MSG(buf);
    agps_cb_start(0);
    page_pm_batch_start();
    gtk_button_set_label(GTK_BUTTON(widget), "Power off");
    gtk_widget_set_sensitive(ptfWidget, false);
}

void
page_main_done()
{
    // Nothing to do at the end of the NMEA epoch?
    // Might be nice to rotate a little roundel or blink an LED, etc.
    // But then we should put it on all pages....
}

#if 0           // Only for stress test future feature.

//      "queue up" the need to stop for polling by the main GTK thread
//      This is to prevent the need to thread-protect everything.
//
// Caution:  called from a timer thread, so all called GTK+ widgets
//      must be protected by gdk_threads_enter() and gdk_threads_leave()

static bool need_to_stop = false;

void
page_main_schedule_stop()
{
    need_to_stop = true;
}

void
page_main_poll_for_stop()
{
    if (!need_to_stop) return;
    need_to_stop = false;
    const char* msg = "stopped";
    if (gllinRunning)
    {
        msg = "stopping...";
        on_clicked_start(0, 0);
    }
    page_stress_msg(msg);
}

//      Stress test calls here to start/stop the GLLIN.
//
// Caution:  called from a timer thread, so all called GTK+ widgets
//      must be protected by gdk_threads_enter() and gdk_threads_leave()

void
page_main_gllin(bool start)
{
    const char* msg = "";
    if (start)
    {
        if (gllinRunning)
        {
            msg = "Running";
        }
        else
        {
            msg = "starting...";
            on_clicked_start(0, 0);
        }
    }
    else
    {
        msg = "stopping...";
        page_main_schedule_stop();
        page_main_poll_for_stop();              // remove if g_timeout cb needs mutex
    }
    page_stress_msg(msg);
}

#endif 

void on_clicked_exit(GtkWidget* widget, gpointer data)
{
    g_print("EXIT button\n");
    agps_stop();
    gtk_main_quit();
}

void on_clicked_reset(GtkWidget* widget, gpointer data)
{
	char c = GPOINTER_TO_INT(data);

	//printf("performing %c reset\n", c);

	clearStats();
	gllin_reset(c);
}

static gboolean
expose_main_event(GtkWidget* w, GdkEventExpose* event)
{
    notebook_expose("MN");
    return FALSE;
}

GtkWidget*
page_main_create(GtkWidget* window)
{
    GtkWidget* label;
    GtkWidget* table;

    //--------------------------------------------------------------------------
    //          table
    //--------------------------------------------------------------------------
    table = gtk_table_new(14, 2, TRUE);
    gtk_signal_connect(GTK_OBJECT(table), "expose_event",
                    G_CALLBACK(expose_main_event), table);

    //--------------------------------------------------------------------------
    //          label
    //--------------------------------------------------------------------------
    label = gtk_label_new(page_main_omgui_version());
    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE (table), label, 0, 2, 0, 1);
    gtk_widget_show(label);

    //--------------------------------------------------------------------------
    //          label
    //
    //  All text from the "main" page goes to this agpsLabel
    //--------------------------------------------------------------------------
    label = gtk_label_new("Press START to begin GPS");
    agpsLabel = GTK_LABEL(label);
    gtk_label_set_justify(agpsLabel, GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 2, 1, 13);
    gtk_widget_show(label);

    {
	    GtkWidget *box, *button;

	    box = gtk_hbox_new (FALSE, 0);

	    button = gtk_button_new_with_label("Cold reset");
	    g_signal_connect(G_OBJECT(button), "clicked",
			    G_CALLBACK(on_clicked_reset), GINT_TO_POINTER('c'));
	    gtk_box_pack_start_defaults(GTK_BOX(box), button);

	    button = gtk_button_new_with_label("Warm reset");
	    g_signal_connect(G_OBJECT(button), "clicked",
			    G_CALLBACK(on_clicked_reset), GINT_TO_POINTER('w'));
	    gtk_box_pack_start_defaults(GTK_BOX(box), button);

	    button = gtk_button_new_with_label("Hot reset");
	    g_signal_connect(G_OBJECT(button), "clicked",
			    G_CALLBACK(on_clicked_reset), GINT_TO_POINTER('h'));
	    gtk_box_pack_start_defaults(GTK_BOX(box), button);

	    gtk_widget_show_all(box);

	    gtk_table_attach_defaults(GTK_TABLE (table), box, 0, 2, 12, 13);
    }
    //--------------------------------------------------------------------------
    //          range controls
    //--------------------------------------------------------------------------
    create_range_controls(window, table, false);

    //--------------------------------------------------------------------------
    //          box for all the buttons
    //--------------------------------------------------------------------------
    GtkWidget* box = gtk_hbox_new (FALSE, 0);

    //--------------------------------------------------------------------------
    //          start periodic button
    //--------------------------------------------------------------------------
    GtkWidget* button = gtk_button_new_with_label("Power on");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_start), GINT_TO_POINTER(0));
    gtk_widget_show(button);
    startWidget = button;
    gtk_box_pack_start(GTK_BOX (box), button, TRUE, TRUE, 1);

    //--------------------------------------------------------------------------
    //          PTF button
    //--------------------------------------------------------------------------
    button = gtk_button_new_with_label("One Fix");
    g_signal_connect(G_OBJECT(button), "clicked",
                     G_CALLBACK(on_clicked_start), GINT_TO_POINTER(1));
    gtk_widget_show(button);
    ptfWidget = button;
    gtk_box_pack_start(GTK_BOX (box), button, TRUE, TRUE, 1);

    //--------------------------------------------------------------------------
    //          exit button
    //--------------------------------------------------------------------------
    button = gtk_button_new_with_label("Exit");
    g_signal_connect(G_OBJECT(button), "clicked",
                    G_CALLBACK(on_clicked_exit), NULL);
    gtk_widget_show(button);
    gtk_box_pack_start(GTK_BOX (box), button, TRUE, TRUE, 1);

    gtk_table_attach_defaults(GTK_TABLE (table), box, 0, 2, 13, 14);
    gtk_widget_show(box);

    gtk_widget_show(table);
    return table;
    test_name[0] = 0;
}

extern void pm_report(FILE* f, bool stress);

void
page_main_report(FILE* f)
{
    fprintf(f, "GPS is %s\n", (gllinRunning) ? "Running": "Stopped");
    pm_report(f, false);
    fprintf(f, "%s\n", gtk_label_get_text(agpsLabel));
}
